/***************************************************************************/
/*                                                                         */
/* FILE NAME : MyGAfrGP.c                                                  */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/*
  Program for graph partitioning using GA's, written by Dr. Ching-Tai Chiu 
*/

#include "ga.h"
#include <stdio.h>

extern void generate_initial_set();
extern void generate_graph();
extern double calc_pop_fitness(), calc_fitness();
extern void reproduction();
extern void rescale();
int evol, spop, besti;
double zs[MAX_POP];        /* Fitness value of this solution */ 
double w[MAX_ND][MAX_ND];  /* Weights between graph nodes */
double wS[MAX_ND][MAX_ND]; /* Starting Weights between graph nodes */
double MainFitnessValue;
float p[MAX_POP];
int nodecount;
int bstr[MAX_POP][MAX_ND], bestsol[MAX_ND];
FILE *fp;

DebugPrint(str)
char *str;
{ /*printf("\n %s \n",str);
  */
  return;
}

/*
*  Normalize the weight values such that sum of weights is equal to 1.
*/
void Normalize(nodenum)
int nodenum;
{  int i,j;
   double total = 0.0;

   for(i=0;i<nodenum;i++) 
      for(j=0;j<nodenum;j++) total += w[i][j];

   if(total!=0)
   {  for(i=0;i<nodenum;i++)
        for(j=0;j<nodenum;j++) 
           w[i][j] = w[i][j] / total;
   } 
   else printf("\n In Normalize() total is ZERO \n");
      
}/* end of Normalize() */

/*
* Finds the minimum valued element's index from the double array.
*/
int Find_minimum(darray,size)
double darray[];
int size;
{  int i;
   int min = 0;
   for(i=1;i<size;i++)
      if(darray[min]>darray[i]) min=i;
   return min;
}/* Find_minimum() */

/*
* Prints the population information. Each solution. Each solution's
* fitness value and if it has the minimum fitness value then this solution
* has a double star.
*/
PrintPopulation(fp, gen_num, nodecount, pop_size)
FILE *fp;
int gen_num;
int nodecount;
int pop_size;
{  int i,j; 
   int min;

   min=Find_minimum(zs,pop_size);

   if(gen_num>=0) fprintf(fp,"\n Generation : %d ",gen_num);

   for(i=0;i<pop_size;i++)
   {  fprintf(fp,"\n Solution %3d -->",i); 
      for(j=0;j<nodecount;j++) fprintf(fp, "%d", bstr[i][j]);
      fprintf(fp,"  %8.4f ",zs[i]);
      if((i==min)||(zs[i]==zs[min])) fprintf(fp," ** ");
   }
   fprintf(fp,"\n");
}/* end of PrintPopulation() */

solveProblemOnce(evol, nodecount, spop , fp)
int evol;
int nodecount;
int spop;
FILE *fp;
{
  int i, j, worsti, count0, count1;
  double avr_fit, zmax;
  long int clock;
  char mfx[30], mfy[30];

  DebugPrint(" I generated GRAPH "); 

  clock = time(0);
  generate_initial_set();   /* Randomly find the weight values */
  /*Normalize(nodecount);     Normalize the weight values */
  avr_fit = calc_pop_fitness();  /* Calculate the population's fitness */
  rescale();

  /*DebugPrint(" Let's print out what we got so far "); 
  */

  for (i=0; i<spop; i++) {
    fprintf(fp, "\n%d    %6.2f  %6.3f --- ", i, zs[i], p[i]);
    for (j=0; j<nodecount; j++) fprintf(fp, "%d", bstr[i][j]);
  }

  fprintf(fp, "\nAverage fitness = %8.4f    Min[%d]=%6.3f\n", 
	  avr_fit, besti, zs[besti]);

  fprintf(fp, " %d\n", 0);
  fprintf(fp, " %5.0f\n", zs[besti]);

  /* HERE main generation starts. */

 printf("Generation : ");

  for (i=0; i<evol; i++) 
  {
    printf(" %d ",i);

    /*
    PrintPopulation(fp, i, nodecount, spop);
    */

    reproduction();

/*    zmax = zs[0];
    worsti = 0;
    for (j=0; j<spop; j++) 
      if (zmax < zs[j]) {
	worsti = j;
	zmax = zs[j];
      }
    for (j=0; j<nodecount; j++)	bstr[worsti][j] = bestsol[j];
*/
    avr_fit = calc_pop_fitness();

    /*
    printf("Average fitness at %d is %8.3f ",i,avr_fit);
    DebugPrint("EVOL2");
    */

    rescale();

    /*
    DebugPrint("EVOL3");

    printf("\nGeneration %4d: average fitness = %8.4f    Min[%d]=%6.3f\n", 
	   i+1, avr_fit, besti, zs[besti]);
    */

    /*
    fprintf(fp, " %d\n", i+1);
    fprintf(fp, " %5.0f\n", zs[besti]);

    if ((i % 10) == 0)
      fprintf(fp, "\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);   
    */


    if (drand48()<0.02 || (avr_fit - zs[besti]) < 0.005) {

      mutation();

      avr_fit = calc_pop_fitness();

      rescale();

/*
      fprintf(fp, "\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);
      fprintf(fp, " %d\n", i+1);
      fprintf(fp, " %5.0f\n", zs[besti]);
    */ 

      /*
      printf("\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);
      printf( " %d\n", i+1);
      printf( " %5.0f\n", zs[besti]);
      */


/*      zmax = zs[0];
      worsti = 0;
      for (j=0; j<spop; j++) 
	if (zmax < zs[j]) {
	  worsti = j;
	  zmax = zs[j];
	}
      for (j=0; j<nodecount; j++) bstr[worsti][j] = bestsol[j];  */


    }
  }

  clock = time(0) - clock;
  for (i=0; i<spop; i++) {
    fprintf(fp, "\n%d    %6.2f  %6.3f --- ", i, zs[i], p[i]);
    for (j=0; j<nodecount; j++) fprintf(fp, "%d", bstr[i][j]);
  }

  fprintf(fp, "\nNodes = %d\nPopulation = %d\nGenerations = %d",
	  nodecount, spop, evol);

  fprintf(fp, "\n\nBest cost = %8.2f -> ", calc_fitness(bstr[besti]));
  count0 = count1 = 0;
  for (i=0; i<nodecount; i++) 
    {
      fprintf(fp, "%d", bstr[besti][i]);
      if (bstr[besti][i] == 0)
	count0++;
      else
	count1++;
    }
  fprintf(fp, "\n\n%d nodes in 0 set\n%d nodes in 1 set\n", count0, count1);

  fprintf(fp, "\nTime = %d\n", clock);
   
  printf("   ---> END \n");

}

/*
*
*/

double MeanOfPopulation(popsize)
int popsize;
{  double ttl = 0.0;
   int i;
   for(i=0;i<popsize;i++)
       ttl = ttl + zs[i];
   return ( ttl / popsize );
}/* MeanOfPopulation() */

/*
*
*/
double VarianceOfPopulation(popsize,mean)
int popsize;
double mean;
{  double ttl;
   int i;
   for(i=0;i<popsize;i++)
      ttl = ttl + square((zs[i]-mean));
   return ( ttl );
}/* VarianceOfPopulation() */

/*
*
*/
double solveOneProblem(evol, nodecount, spop , fp, NumOfTimes)
int evol;
int nodecount;
int spop;
FILE *fp;
int NumOfTimes;
{
  int i, j, bidx;
  double avr_fit, zmax;
  long int clock;
  char mfx[30], mfy[30];
  double TheBestVariance, TheBestMean, TheBestFittness;

  solveProblemOnce(evol, nodecount, spop , fp); 
  TheBestFittness = zs[Find_minimum(zs,spop)];
  TheBestMean=MeanOfPopulation(spop);
  TheBestVariance=VarianceOfPopulation(spop,TheBestMean);

  for(i=1;i<NumOfTimes;i++)
  {  
      solveProblemOnce(evol, nodecount, spop , fp); 
      bidx = Find_minimum(zs,spop);
      if(zs[bidx]<TheBestFittness)
      {   TheBestFittness = zs[bidx];
	  for(j=0;j<nodecount;j++) bestsol[j] = bstr[bidx][j];
          TheBestMean = MeanOfPopulation(spop);
          TheBestVariance = VarianceOfPopulation(spop,TheBestMean);
      }/* end of if */

  }/* end of for */

  /* Produce report */
  fprintf(fp,"\n The Solution :");
  for(i=0;i<nodecount;i++) fprintf(fp,"%d",bestsol[i]);
  fprintf(fp,"\n");
  fprintf(fp,"\n The Fittness : %f ",TheBestFittness);
  fprintf(fp,"\n The Mean of the population : %f ",TheBestMean);
  fprintf(fp,"\n The Variance of the population : %f \n",TheBestVariance);

  return ( TheBestFittness );

}/* solveOneProblem() */

Show_Graph(nodecount)
int nodecount;
{  int k,m;

   /* prints the first spop-1 rows of the weight matrix */
   for(k=0;k<(spop-1);k++)
   {   for(m=0;m<nodecount;m++) printf("%9.4f ",w[k][m]);
       printf("\n");
   }/* end of for */

   /* prints the last row of the weight matrix */
   for(m=0;m<nodecount;m++) printf("%9.4f ",w[k][m]);

   /* prints the fitness value of this problem
   printf("%9.4f \n",zs[Find_minimum(zs,spop)]); */

}/* Show_Graph() */

/*
*
*
*/
read_weights(fp,k)
FILE *fp;
int k;
{  int i,j;
   int EndOfFile = 1;
   float fbuf;
   i=j=0;
   while((i<k)&&(EndOfFile))
   {  j=0;
      while((j<k)&&(EndOfFile))
      { if(j<(k-1))
        { if(fscanf(fp,"%f ",&fbuf)==EOF) EndOfFile = 0;
          else w[i][j] = fbuf;
        }
        else
         { if(fscanf(fp,"%f \n",&fbuf)==EOF) EndOfFile = 0;
           else w[i][j] = fbuf;
         }
        j++;
      }
      i++;
   }
   printf("\n read_weights i = %d \n",i);
   Show_Graph(k);
   return(EndOfFile);
}/* read_weights() */


Print_Graph(nodecount,fp)
int nodecount;
FILE *fp;
{  int k,m;

   /* prints the first spop-1 rows of the weight matrix */
   for(k=0;k<(spop-1);k++)
   {   for(m=0;m<nodecount;m++) fprintf(fp,"%9.4f ",w[k][m]);
       fprintf(fp,"\n");
   }/* end of for */

   /* prints the last row of the weight matrix */
   for(m=0;m<nodecount;m++) fprintf(fp,"%9.4f ",w[k][m]);

   /* prints the fitness value of this problem */
   fprintf(fp,"%9.4f \n",zs[Find_minimum(zs,spop)]);

}/* Print_Graph() */

       
domain(outFile, NumberOfProblems,flag,QuesFileName)
char *outFile;
int NumberOfProblems;
int flag;
char *QuesFileName;
{
  int i, j, worsti, count0, count1;
  double avr_fit, zmax;
  long int clock;
  FILE *mpx, *mpy;
  char mfx[30], mfy[30];
  char GrFileName[30];
  int k, m;
  FILE *fpResults;
  FILE *fpStatistics;
  FILE *QuesFile;
  double VarFitness,VarWeights;
  double diff;
  double FValue;

/*
  if (argc==1)
    {
      printf("Usage:\n");
      printf("num_nodes population epochs outputfile\n");
      exit(0);
    }
  sscanf(argv[1],"%d", &nodecount);
  sscanf(argv[2], "%d", &spop);
  sscanf(argv[3], "%d", &evol);

  if ((fp = fopen(argv[4], "w+t")) == NULL)
    {
      printf("Could not open output file\n");
      exit(0);
    }
*/
  if ((fp = fopen(outFile, "w+t")) == NULL)
    {
      printf("Could not open output file\n");
      return;
    }

  if ((fpStatistics = fopen("Stat", "w")) == NULL)
    {
      printf("Could not open output file\n");
      exit(0);
    }


  if ((fpResults = fopen("Results", "w")) == NULL)
    {
      printf("Could not open output file\n");
      exit(0);
    }
  DebugPrint("I opened the output file ");

  rewind(fp);

  /*
  strcpy(mfx, argv[4]);
  strcpy(mfy, argv[4]);
  */
  strcpy(mfx, "x.mat");
  strcpy(mfy, "y.mat");
  mpx = fopen(mfx, "w+t");
  mpy = fopen(mfy, "w+t");

  /* Write the number of nodes in the graph */

  fprintf(fpStatistics,"%d \n",nodecount);

  sprintf(GrFileName,"Graph%d",i);

  if(flag)
  {  if((QuesFile=fopen(QuesFileName,"r"))!=NULL) 
     {  read_weights(QuesFile,nodecount); 
	printf("\n Read is finished \n");
	fclose(QuesFile);
     }
     else generate_graph_write(GrFileName);
  }
  else generate_graph_write(GrFileName);

  for(i=0;i<NumberOfProblems;i++)
  {
     printf("\n Problem %d \n",i);

     if(!i)
     {  sprintf(GrFileName,"Graph%d",i);

        generate_graph_write(GrFileName);
     }

     FValue = solveOneProblem(evol, nodecount, spop , fp, 4);

     /*printf("\n After SolveOneProblem \n");*/

     Print_Graph(nodecount,fpStatistics);

     /* prints the fitness value of this problem */
     fprintf(fp,"%9.4f \n",FValue);

     /*printf("\n After Print_Graph \n"); */

  }/* end of for */

  fclose(mpx);
  fclose(mpy);
  fclose(fp);
  fclose(fpStatistics);
  fclose(fpResults);

}/* end of domain() */

       
main(argc, argv)
int argc;
char *argv[];
{
  char InFileName[30];
  int NumberOfProblems=40; /* Default */

  if (argc>4)
  {
     sscanf(argv[1],"%d", &nodecount); /*  nodecount = atoi(argv[1]);  */
     sscanf(argv[2], "%d", &spop);     /*  spop = atoi(argv[2]);  */
     sscanf(argv[3], "%d", &evol);     /*  evol = atoi(argv[3]);  */

     if((argc==6)&&(isdigit(argv[5][0])))
              sscanf(argv[5],"%d",&NumberOfProblems);
     if(argc==7)
     {  sprintf(InFileName,"%s",argv[6]);
        InFileName[29] = NULL;
	printf("\n I need to read from the %s \n",InFileName);
     }

     

     /* Produce Report */
     printf("\n Number of Nodes .....: %6d",nodecount);
     printf("\n Population Size .....: %6d",spop);
     printf("\n Maximum epoch number : %6d",evol);
     printf("\n Number of Problems ..: %6d \n",NumberOfProblems);

     domain(argv[4],NumberOfProblems,(argc==7),InFileName);

  }
  else
    {
      printf("Usage:\n");
      printf("num_nodes population epochs outputfile\n");
      exit(0);
    }

}/* end of main */
